from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from krita import *
import ctypes
from ctypes import cdll
import os
import sys
import platform
import struct
import math

KI = Krita.instance()

# Load native library
dllext = '.so'
if platform.system() == 'Windows':
    dllext = '.dll'
DLL = cdll.LoadLibrary(os.path.dirname(os.path.abspath(__file__)) + os.path.sep + "view3d" + dllext)

# DLL constants
DISTANCE_GRADIENT=0
DIRECTION_GRADIENT=1

# Callback for getting OpenGL procedure addresses from the Qt OpenGL context
getProcAddressCallbackContext = [None]
@ctypes.CFUNCTYPE(ctypes.c_longlong, ctypes.c_char_p)
def getProcAddressCallback(name):
    return int(getProcAddressCallbackContext[0].getProcAddress(name))

class View3DViewport(QOpenGLWidget):
    _gl = None
    _initFailed = False
    _liveCheckbox = None
    _mesh = 0
    _tgtx = 0
    _tgty = 0
    _tgtz = 0
    _rotx = 0
    _roty = 0
    _camd = 10
    _camRot = False
    _camPan = False
    _camZoom = False
    _viewSelection = False
    _viewOrthographic = False
    _viewBgColor = QColor(26, 51, 76)
    _doc = None
    _mx = 0
    _my = 0
    _tiImage = None
    _tiImageOwnerDoc = None
    _tiImageDoc = None
    _fiLastLayer = None
    _periodicUpdates = True

    def __init__(self, parent=None):
        super(View3DViewport, self).__init__(parent)
        fmt = QSurfaceFormat()
        fmt.setRedBufferSize(8)
        fmt.setGreenBufferSize(8)
        fmt.setBlueBufferSize(8)
        fmt.setAlphaBufferSize(8)
        fmt.setSamples(0)
        fmt.setSwapBehavior(QSurfaceFormat.DoubleBuffer)
        fmt.setSwapInterval(0)
        fmt.setMajorVersion(2)
        fmt.setMinorVersion(1)
        fmt.setProfile(QSurfaceFormat.CompatibilityProfile)
        self.setFormat(fmt)
        self._timer = QTimer(self)
        self._timer.timeout.connect(self.timerTick)
        self._timer.start(33)
        self._toggleFilteringAction = QAction("Filtering")
        self._toggleFilteringAction.triggered.connect(self.toggleFiltering)
        self._toggleFilteringAction.setCheckable(True)
        self._toggleFilteringAction.setChecked(False)
        self._toggleShadingAction = QAction("Shading")
        self._toggleShadingAction.triggered.connect(self.toggleShading)
        self._toggleShadingAction.setCheckable(True)
        self._toggleShadingAction.setChecked(True)
        self._toggleWireframeAction = QAction("Wireframe Overlay")
        self._toggleWireframeAction.triggered.connect(self.toggleWireframe)
        self._toggleWireframeAction.setCheckable(True)
        self._toggleBackfacesAction = QAction("Backfaces")
        self._toggleBackfacesAction.triggered.connect(self.toggleBackfaces)
        self._toggleBackfacesAction.setCheckable(True)
        self._toggleBackfacesAction.setChecked(True)
        self._toggleOrthographicAction = QAction("Orthographic")
        self._toggleOrthographicAction.triggered.connect(self.toggleOrthographic)
        self._toggleOrthographicAction.setCheckable(True)
        self._toggleSelectionAction = QAction("Selection")
        self._toggleSelectionAction.triggered.connect(self.toggleSelection)
        self._toggleSelectionAction.setCheckable(True)
        self._resetViewAction = QAction("Reset View")
        self._resetViewAction.triggered.connect(self.resetView)
        self._alignCanvasToViewAction = QAction("Align Canvas to View")
        self._alignCanvasToViewAction.triggered.connect(self.alignCanvasToView)
        self._backgroundColorAction = QAction("Background Color...")
        self._backgroundColorAction.triggered.connect(self.backgroundColor)
        self._toImageAction = QAction("Create Image from View")
        self._toImageAction.triggered.connect(self.toImage)
        self._fromImageAction = QAction("Backproject Last Created Image")
        self._fromImageAction.triggered.connect(self.fromImage)
        self._createMaskAction = QAction("Create Mask from View")
        self._createMaskAction.triggered.connect(self.createMask)
        self._createTexCoordMaskAction = QAction("Create Texture Coordinate Mask")
        self._createTexCoordMaskAction.triggered.connect(self.createTexCoordMask)
        self._createTexCoordWireframeAction = QAction("Create Texture Coordinate Wireframe")
        self._createTexCoordWireframeAction.triggered.connect(self.createTexCoordWireframe)
        self._triplanarPasteAction = QAction("Triplanar Paste...")
        self._triplanarPasteAction.triggered.connect(self.triplanarPaste)
        self._createGradientAction = QAction("Create Gradient...")
        self._createGradientAction.triggered.connect(self.createGradient)
        self._selectionAsAlphaMaskAction = QAction("Selection as Alpha Mask")
        self._selectionAsAlphaMaskAction.setCheckable(True)
        self._selectionAsAlphaMaskAction.setChecked(True)
        self._setTextureToCurrentDocumentAction = QAction("Set Texture to Current Document")
        self._setTextureToCurrentDocumentAction.triggered.connect(self.setTextureToCurrentDocument)
        self._setTextureToAnyDocumentAction = QAction("Set Texture to Any Document")
        self._setTextureToAnyDocumentAction.triggered.connect(self.setTextureToAnyDocument)
        self._scalingAt100PercentAction = QAction("Scaling at 100%")
        self._scalingAt100PercentAction.triggered.connect(self.scalingAt100Percent)
        self._scalingAt50PercentAction = QAction("Scaling at 50%")
        self._scalingAt50PercentAction.triggered.connect(self.scalingAt50Percent)
        self._scalingAt25PercentAction = QAction("Scaling at 25%")
        self._scalingAt25PercentAction.triggered.connect(self.scalingAt25Percent)
        self._scalingAt125PercentAction = QAction("Scaling at 12.5%")
        self._scalingAt125PercentAction.triggered.connect(self.scalingAt125Percent)
        self._usePaletteAction = QAction("Use palette...")
        self._usePaletteAction.triggered.connect(self.usePalette)
        self._openMeshAction = QAction("Open Mesh...")
        self._openMeshAction.triggered.connect(self.openMesh)
        self._loadCubeAction = QAction("Load Cube")
        self._loadCubeAction.triggered.connect(self.loadCube)
        self._loadQuadAction = QAction("Load Quad")
        self._loadQuadAction.triggered.connect(self.loadQuad)
        self._openRecentAction = [None]*10
        for i in range(10):
            self._openRecentAction[i] = QAction(KI.readSetting("view3dplugin", "recent" + str(i), ""))
            self._openRecentAction[i].triggered.connect(self.openRecent)
        self._clearRecentAction = QAction("Clear Recent Meshes")
        self._clearRecentAction.triggered.connect(self.clearRecent)

    def msgBox(self, msg):
        box = QMessageBox(self)
        box.setText(msg)
        box.setWindowTitle('View3D Message')
        box.setStandardButtons(QMessageBox.Ok)
        box.exec()
        box.deleteLater()

    def selectItemFromListBox(self, title, text, items, default=0):
        dlg = QDialog(self)
        dlg.setWindowTitle(title)
        layout = QFormLayout(dlg)
        layout.addRow(QLabel(text, dlg))
        listBox = QListWidget(dlg)
        for i in items:
            listBox.addItem(i)
        listBox.setCurrentRow(default)
        layout.addRow(listBox)
        box = QHBoxLayout()
        box.addStretch(1)
        btn = QPushButton('&OK', dlg)
        btn.setDefault(True)
        btn.clicked.connect(dlg.accept)
        box.addWidget(btn)
        btn = QPushButton('&Cancel', dlg)
        btn.clicked.connect(dlg.reject)
        box.addWidget(btn)
        layout.addRow(box)
        if dlg.exec() != QDialog.Accepted:
            dlg.deleteLater()
            return -1
        dlg.deleteLater()
        return listBox.currentRow()

    def getDoc(self):
        if self._doc != None:
            return self._doc
        return KI.activeDocument()

    def applySelectionAsAlphaMask(self, pixels):
        if not self._selectionAsAlphaMaskAction.isChecked():
            return
        doc = self.getDoc()
        if doc == None:
            return
        sel = doc.selection()
        if sel == None:
            return
        sela = bytes(sel.pixelData(0, 0, doc.width(), doc.height()))
        DLL.applyAlphaMask(sela, pixels, doc.width(), doc.height())

    def updateTexture(self):
        if self._initFailed:
            return
        doc = self.getDoc()
        if doc and self._mesh != 0:
            img = doc.projection(0, 0, doc.width(), doc.height())
            DLL.setMeshTextureData(self._mesh, 0, doc.width(), doc.height(), ctypes.c_void_p(int(img.constBits())))
            if self._viewSelection:
                sel = doc.selection()
                if sel == None or sel.width() < 1 or sel.height() < 1:
                    DLL.setSelection(0)
                    DLL.setMeshSelectionData(self._mesh, 0, 0, 0, 0, None)
                else:
                    bits = bytes(sel.pixelData(sel.x(), sel.y(), sel.width(), sel.height()))
                    DLL.setSelection(1)
                    DLL.setMeshSelectionData(self._mesh, sel.x(), sel.y(), sel.width(), sel.height(), bits)
            else:
                DLL.setSelection(0)
            self.repaint()

    def timerTick(self):
        if self._initFailed:
            self._timer.stop()
            box = QMessageBox(self)
            box.setText("Failed to initialize OpenGL, make sure Krita is using the OpenGL display backend.")
            box.setWindowTitle('View3D Message')
            box.setStandardButtons(QMessageBox.Ok)
            box.exec()
            self.setVisible(False)
            QLabel('No OpenGL', self.parent())
            return
        if self._periodicUpdates and self.isVisible():
            self.updateTexture()
        if self._doc != None or self._tiImageOwnerDoc != None:
            docFound = False
            iodFound = False
            for d in KI.documents():
                if d == self._doc:
                    docFound = True
                elif d == self._tiImageOwnerDoc:
                    iodFound = True
                if docFound and iodFound:
                    break
            if not docFound:
                self._doc = None
                self._fiLastLayer = None
            if not iodFound:
                self._tiImage = None
                self._tiImageDoc = None
                self._tiImageOwnerDoc = None

    def texCoordAtXY(self, x, y):
        s = ctypes.c_int()
        t = ctypes.c_int()
        if DLL.texCoordAtXY(x, y, ctypes.byref(s), ctypes.byref(t)) == 0:
            return (s.value, t.value)
        return (-1, -1)

    def plotPoint(self, x, y):
        doc = self.getDoc()
        if not doc:
            return
        x, y = self.texCoordAtXY(x, y)
        if x < 0 or y < 0 or x >= doc.width() or y >= doc.height():
            return
        doc.activeNode().setPixelData(QByteArray(b'\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF'), x - 1, y - 1, 3, 3)
        doc.refreshProjection()

    def mouseMoveEvent(self, event):
        super(View3DViewport, self).mouseMoveEvent(event)
        if self._camRot:
            self._rotx += event.globalY() - self._my
            self._roty += event.globalX() - self._mx
            self._mx = event.globalX()
            self._my = event.globalY()
            self.repaint()
        if self._camPan:
            dx = -0.015*(event.globalX() - self._mx)
            dy = 0.015*(event.globalY() - self._my)
            rx = math.radians(self._rotx)
            ry = math.radians(self._roty)
            self._tgtx += math.cos(ry)*dx
            self._tgty += math.cos(rx)*dy
            self._tgtz += math.sin(ry)*dx - math.sin(rx)*dy
            self._mx = event.globalX()
            self._my = event.globalY()
            self.repaint()
        if self._camZoom:
            dy = 0.05*(event.globalY() - self._my)
            ry = math.radians(self._roty)
            self._camd = max(0.0, self._camd + dy)
            self._my = event.globalY()
            self.repaint()

    def mousePressEvent(self, event):
        super(View3DViewport, self).mousePressEvent(event)
        if self._tiImageDoc != None and self._tiImageDoc == KI.activeDocument():
            self._mx = event.globalX()
            self._my = event.globalY()
            return
        if event.button() == Qt.LeftButton:
            if event.modifiers() == Qt.AltModifier:
                self.plotPoint(event.x(), event.y())
                return
            self._camRot = True
            self._mx = event.globalX()
            self._my = event.globalY()
            return
        if event.button() == Qt.MiddleButton:
            self._camPan = True
            self._mx = event.globalX()
            self._my = event.globalY()
            return
        if event.button() == Qt.RightButton:
            self._camZoom = True
            self._mx = event.globalX()
            self._my = event.globalY()
            return

    def mouseReleaseEvent(self, event):
        super(View3DViewport, self).mouseReleaseEvent(event)
        DLL.markViewDataDirty()
        self.repaint()
        if event.button() == Qt.LeftButton:
            self._camRot = False
            self._viewRot = False
            return
        if event.button() == Qt.MiddleButton:
            self._camPan = False
            return
        if event.button() == Qt.RightButton:
            self._camZoom = False
            if self._mx == event.globalX() and self._my == event.globalY():
                self.showPopupMenu()
            return

    def wheelEvent(self, event):
        self._camd = max(0.0, self._camd - event.angleDelta().y()/64.0)
        self.repaint()

    def showPopupMenu(self):
        # Note: assumes _mx,_my are set to mouse coordinates
        menu = QMenu(self)
        if self._tiImageDoc != None and self._tiImageDoc == KI.activeDocument():
            menu.addAction(self._fromImageAction)
            menu.addSeparator()
        menu.addAction(self._toggleFilteringAction)
        menu.addAction(self._toggleShadingAction)
        menu.addAction(self._toggleWireframeAction)
        menu.addAction(self._toggleBackfacesAction)
        menu.addAction(self._toggleOrthographicAction)
        menu.addAction(self._toggleSelectionAction)
        menu.addSeparator()
        if self._tiImageDoc == None or self._tiImageDoc != KI.activeDocument():
            menu.addAction(self._resetViewAction)
            menu.addAction(self._alignCanvasToViewAction)
        menu.addAction(self._backgroundColorAction)
        menu.addSeparator()
        images = menu.addMenu("Images")
        if self._tiImageDoc == None or self._tiImageDoc != KI.activeDocument():
            images.addAction(self._toImageAction)
        images.addAction(self._fromImageAction)
        images.addSeparator()
        images.addAction(self._createMaskAction)
        images.addAction(self._createTexCoordMaskAction)
        images.addAction(self._createTexCoordWireframeAction)
        images.addSeparator()
        images.addAction(self._triplanarPasteAction)
        images.addAction(self._createGradientAction)
        images.addSeparator()
        images.addAction(self._selectionAsAlphaMaskAction)
        menu.addSeparator()
        texmenu = menu.addMenu("Texture")
        texmenu.addAction(self._setTextureToCurrentDocumentAction)
        texmenu.addAction(self._setTextureToAnyDocumentAction)
        texmodmenu = menu.addMenu("Texture modifiers")
        texmodmenu.addAction(self._scalingAt100PercentAction)
        texmodmenu.addAction(self._scalingAt50PercentAction)
        texmodmenu.addAction(self._scalingAt25PercentAction)
        texmodmenu.addAction(self._scalingAt125PercentAction)
        texmodmenu.addSeparator()
        texmodmenu.addAction(self._usePaletteAction)
        menu.addSeparator()
        openmenu = menu.addMenu("Mesh")
        openmenu.addAction(self._openMeshAction)
        openmenu.addSeparator()
        openmenu.addAction(self._loadCubeAction)
        openmenu.addAction(self._loadQuadAction)
        openmenu.addSeparator()
        anyRecent = False
        for i in range(10):
            if len(self._openRecentAction[i].text()) > 0:
                openmenu.addAction(self._openRecentAction[i])
                anyRecent = True
        if anyRecent:
            openmenu.addSeparator()
            openmenu.addAction(self._clearRecentAction)
        menu.popup(QPoint(self._mx, self._my))

    def resetView(self):
        self._tgtx = 0
        self._tgty = 0
        self._tgtz = 0
        self._rotx = 0
        self._roty = 0
        self._camd = 10
        DLL.markViewDataDirty()
        self.repaint()

    def alignCanvasToView(self):
        # Note: assumes _mx,_my are set to mouse coordinates
        mousepos = self.mapFromGlobal(QPoint(self._mx, self._my))
        origin = self.texCoordAtXY(mousepos.x(), mousepos.y())
        if origin[0] < 0:
            return
        found = False
        todownx = 0
        todowny = 0
        lastx = -1000000
        lasty = -1000000
        for i in range(0, 16):
            down = self.texCoordAtXY(mousepos.x(), mousepos.y() + i)
            if (down[0] != -1) and ((down[0] != origin[0]) or (down[1] != origin[1])) and (down[0] != lastx) and (down[1] != lasty):
                lastx = down[0]
                lasty = down[1]
                todownlen = math.sqrt((down[0] - origin[0])**2 + (down[1] - origin[1])**2)
                if todownlen < 10:
                    todownx += float(down[0] - origin[0])
                    todowny += float(down[1] - origin[1])
                    found = True
        if found:
            todownlen = math.sqrt(todownx**2 + todowny**2)
            todownx /= todownlen
            todowny /= todownlen
            dot = todowny
            det = todownx
            a = math.atan2(det, dot)*180.0/3.1415;
            win = KI.activeWindow()
            if win != None:
                view = win.activeView()
                if view != None:
                    canvas = view.canvas()
                    if canvas != None:
                        restoreOptions = False
                        qwin = Krita.instance().activeWindow().qwindow()
                        if qwin != None:
                            mdi = qwin.findChild(QMdiArea)
                            if mdi != None:
                                scroll = mdi.currentSubWindow().findChild(QAbstractScrollArea)
                                if scroll != None:
                                    viewport = scroll.viewport()
                                    if viewport != None:
                                        restoreOptions = True
                                        zoom = canvas.zoomLevel()
                                        mirror = canvas.mirror()
                                        wrap = canvas.wrapAroundMode()
                                        canvas.setWrapAroundMode(False)
                                        canvas.setMirror(False)
                                        canvas.resetZoom()
                                        canvas.resetRotation()
                                        scroll.horizontalScrollBar().setValue(origin[0] - viewport.width()//2)
                                        scroll.verticalScrollBar().setValue(origin[1] - viewport.height()//2)
                        canvas.setRotation(a)
                        if restoreOptions:
                            canvas.setZoomLevel(zoom)
                            canvas.setMirror(mirror)
                            canvas.setWrapAroundMode(wrap)

    def backgroundColor(self):
        newcol = QColorDialog.getColor(self._viewBgColor, self, "Background Color", QColorDialog.DontUseNativeDialog)
        if newcol.isValid():
            self._viewBgColor = newcol
            self.repaint()

    def toggleShading(self):
        DLL.setShading(self._toggleShadingAction.isChecked())
        self.repaint()

    def toggleFiltering(self):
        DLL.setFiltering(self._toggleFilteringAction.isChecked())
        self.repaint()

    def toggleWireframe(self):
        DLL.setWireframe(self._toggleWireframeAction.isChecked())
        self.repaint()

    def toggleBackfaces(self):
        DLL.setBackfaces(self._toggleBackfacesAction.isChecked())
        self.repaint()

    def toggleOrthographic(self):
        self._viewOrthographic = not self._viewOrthographic
        self.repaint()

    def toggleSelection(self):
        self._viewSelection = not self._viewSelection
        self.updateTexture()
        self.repaint()

    def saveRecents(self):
        for i in range(10):
            QAction(KI.writeSetting("view3dplugin", "recent" + str(i), self._openRecentAction[i].text()))

    def addToRecents(self, path):
        for i in range(10):
            if self._openRecentAction[i].text() == path:
                return
        for i in reversed(range(9)):
            self._openRecentAction[i + 1].setText(self._openRecentAction[i].text())
        self._openRecentAction[0].setText(path)
        self.saveRecents()

    def doOpenMesh(self, path):
        DLL.unloadAllMeshes()
        self._mesh = DLL.loadMeshFile(bytes(path, 'utf-8'))
        if KI.activeDocument() != None:
            self._doc = KI.activeDocument()

    def openMesh(self):
        fn = QFileDialog.getOpenFileName(self, 'Open Mesh', '', 'JTF files (*.jtf)')
        if fn and len(fn[0].strip()) > 0:
            fn = fn[0].strip()
            self.doOpenMesh(fn)
            self.addToRecents(fn)

    def loadCube(self):
        self.doOpenMesh('//dev//null//_Cube_')

    def loadQuad(self):
        self.doOpenMesh('//dev//null//_Quad_')

    def openRecent(self):
        sender = self.sender()
        self.doOpenMesh(sender.text())

    def clearRecent(self):
        for i in range(10):
            self._openRecentAction[i].setText("")
        self.saveRecents()

    def toImage(self):
        gl = self._gl
        self.makeCurrent()
        gl.glClearColor(0.0, 0.0, 0.0, 0.0)
        self.doTheRender(False)
        gl.glClearColor(self._viewBgColor.redF(), self._viewBgColor.greenF(), self._viewBgColor.blueF(), 1.0)
        rgbapixels = gl.glReadPixels(0, 0, self.width(), self.height(), gl.GL_BGRA, gl.GL_FLOAT)
        self.doneCurrent()
        data = bytearray()
        for rgba in rgbapixels:
            data.append(int(rgba*255.0))
        doc = KI.createDocument(self.width(), self.height(), 'Model', 'RGBA', 'U8', '', 100.0)
        self._tiImageOwnerDoc = doc
        bgrow = bytearray()
        for x in range(self.width()):
            bgrow.append(int(self._viewBgColor.blue()))
            bgrow.append(int(self._viewBgColor.green()))
            bgrow.append(int(self._viewBgColor.red()))
            bgrow.append(int(255))
        bgrow = QByteArray(bgrow)
        node = doc.rootNode().childNodes()[0]
        for y in range(self.height()):
            node.setPixelData(bgrow, 0, y, self.width(), 1)
        node = doc.createNode('Render', 'paintlayer')
        node.setOpacity(255)
        for y in range(self.height()):
            node.setPixelData(QByteArray(data[y*self.width()*4:]), 0, self.height() - 1 - y, self.width(), 1)
        doc.rootNode().addChildNode(node, None)
        node = doc.createGroupLayer('Layers to backproject')
        doc.rootNode().addChildNode(node, None)
        node.setOpacity(255)
        node.setPassThroughMode(False)
        node2 = doc.createNode('Paint Here', 'paintlayer')
        node2.setOpacity(255)
        node.addChildNode(node2, None)
        self._tiImage = node
        self._tiImageDoc = doc
        self._fiLastLayer = None
        KI.activeWindow().addView(doc)
        doc.refreshProjection()

    def fromImage(self):
        if self._tiImage == None:
            self.msgBox('Use "To Image" to create an image first')
            return
        doc = self.getDoc()
        if doc == None:
            self.msgBox('You need an active document')
            return
        viewrgba = bytes(self._tiImage.projectionPixelData(0, 0, self.width(), self.height()))
        layerrgba = ctypes.create_string_buffer(doc.width()*doc.height()*4)
        DLL.fromViewToLayerData(viewrgba, layerrgba, doc.width(), doc.height())
        if self._fiLastLayer == None:
            node = doc.createNode('Painted', 'paintlayer')
            node.setOpacity(255)
            doc.rootNode().addChildNode(node, None)
            self._fiLastLayer = node
        else:
            node = self._fiLastLayer
        self.applySelectionAsAlphaMask(layerrgba)
        node.setPixelData(QByteArray(layerrgba.raw), 0, 0, doc.width(), doc.height())
        doc.refreshProjection()

    def createMask(self):
        doc = self.getDoc()
        if doc == None:
            self.msgBox('You need an active document')
            return
        layerrgba = ctypes.create_string_buffer(doc.width()*doc.height()*4)
        DLL.fromViewToWhiteMask(layerrgba, doc.width(), doc.height())
        node = doc.createNode('View Mask', 'paintlayer')
        node.setOpacity(255)
        doc.rootNode().addChildNode(node, None)
        self.applySelectionAsAlphaMask(layerrgba)
        node.setPixelData(QByteArray(layerrgba.raw), 0, 0, doc.width(), doc.height())
        doc.refreshProjection()

    def createTexCoordLayer(self, solid):
        doc = self.getDoc()
        if doc == None:
            self.msgBox('You need an active document')
            return
        if self._mesh == 0:
            self.msgBox("No mesh loaded")
            return
        fc = DLL.getMeshFaceCount(self._mesh)
        if fc == -1:
            self.msgBox("No mesh available")
            return
        img = QImage(QSize(doc.width(), doc.height()), QImage.Format_ARGB32)
        img.fill(QColor(0, 0, 0, 0))
        verts = (ctypes.c_int32*(fc*6))()
        DLL.getMeshTextureCoordinates(self._mesh, ctypes.byref(verts), doc.width(), doc.height(), 0, 0)
        verts = list(verts)
        poly = QPolygon()
        poly.append(QPoint(0, 0))
        poly.append(QPoint(0, 0))
        poly.append(QPoint(0, 0))
        painter = QPainter()
        painter.begin(img)
        painter.setRenderHint(QPainter.Antialiasing, False)
        painter.setPen(Qt.white)
        if solid:
            painter.setBrush(Qt.white)
        for i in range(0, len(verts), 6):
            poly[0] = QPoint(verts[i], verts[i + 1])
            poly[1] = QPoint(verts[i + 2], verts[i + 3])
            poly[2] = QPoint(verts[i + 4], verts[i + 5])
            painter.drawPolygon(poly)
        painter.end()
        bits = img.bits()
        bits.setsize(img.byteCount())
        node = doc.createNode('Tex Coord Mask', 'paintlayer')
        node.setOpacity(255)
        self.applySelectionAsAlphaMask(ctypes.c_void_p(int(bits)))
        newrgba = bytearray(bits.asarray())
        node.setPixelData(QByteArray(newrgba), 0, 0, doc.width(), doc.height())
        doc.rootNode().addChildNode(node, None)
        doc.refreshProjection()

    def createTexCoordMask(self):
        self.createTexCoordLayer(True)

    def createTexCoordWireframe(self):
        self.createTexCoordLayer(False)

    def triplanarPaste(self):
        doc = self.getDoc()
        if doc == None:
            self.msgBox("Need a texture image open")
            return
        img = QGuiApplication.clipboard().image()
        if (img == None) or (img.isNull()):
            self.msgBox("An image must be copied to the clipboard")
            return
        ppu = ctypes.c_float(256)
        DLL.getMeshPixelToUnitRatio(self._mesh, ctypes.byref(ppu))
        ppu = ppu.value
        dlg = QDialog(self)
        dlg.setWindowTitle('Triplanar Paste')
        layout = QFormLayout(dlg)
        smoothSlider = QSlider(Qt.Horizontal, dlg)
        smoothSlider.setMinimum(1000)
        smoothSlider.setMaximum(10000)
        smoothSlider.setValue(3000)
        layout.addRow('Hardness:', smoothSlider)
        ppuSpinBox = QDoubleSpinBox(dlg)
        ppuSpinBox.setMinimum(0.125)
        ppuSpinBox.setMaximum(4096)
        ppuSpinBox.setValue(ppu)
        layout.addRow('Pixels per unit:', ppuSpinBox)
        moatSpinBox = QDoubleSpinBox(dlg)
        moatSpinBox.setValue(4)
        layout.addRow('Moat around geometry:', moatSpinBox)
        xBlendSpinBox = QDoubleSpinBox(dlg)
        xBlendSpinBox.setMinimum(0.0)
        xBlendSpinBox.setMaximum(1.0)
        xBlendSpinBox.setValue(1.0)
        xBlendSpinBox.setSingleStep(0.1)
        yBlendSpinBox = QDoubleSpinBox(dlg)
        yBlendSpinBox.setMinimum(0.0)
        yBlendSpinBox.setMaximum(1.0)
        yBlendSpinBox.setValue(1.0)
        yBlendSpinBox.setSingleStep(0.1)
        zBlendSpinBox = QDoubleSpinBox(dlg)
        zBlendSpinBox.setMinimum(0.0)
        zBlendSpinBox.setMaximum(1.0)
        zBlendSpinBox.setValue(1.0)
        zBlendSpinBox.setSingleStep(0.1)
        box = QHBoxLayout()
        box.addWidget(QLabel('X:'))
        box.addWidget(xBlendSpinBox)
        box.addWidget(QLabel('Y:'))
        box.addWidget(yBlendSpinBox)
        box.addWidget(QLabel('Z:'))
        box.addWidget(zBlendSpinBox)
        layout.addRow('Per-direction blend:', box)
        ignoreNegativeX = QCheckBox('Ignore negative X (for mirrored objects)')
        layout.addRow(ignoreNegativeX)
        box = QHBoxLayout()
        box.addStretch(1)
        btn = QPushButton('&OK', dlg)
        btn.setDefault(True)
        btn.clicked.connect(dlg.accept)
        box.addWidget(btn)
        btn = QPushButton('&Cancel', dlg)
        btn.clicked.connect(dlg.reject)
        box.addWidget(btn)
        layout.addRow(box)
        if dlg.exec() != QDialog.Accepted:
            dlg.deleteLater()
            return
        dlg.deleteLater()
        bits = img.bits()
        bits.setsize(img.byteCount())
        smooth = ctypes.c_float(smoothSlider.value())
        ppu = ctypes.c_float(ppuSpinBox.value())
        moat = ctypes.c_float(moatSpinBox.value())
        xblend = ctypes.c_float(xBlendSpinBox.value())
        yblend = ctypes.c_float(yBlendSpinBox.value())
        zblend = ctypes.c_float(zBlendSpinBox.value())
        ignorenx = 1 if ignoreNegativeX.isChecked() else 0
        layerrgba = ctypes.create_string_buffer(doc.width()*doc.height()*4)
        DLL.triplanarBlend(self._mesh, layerrgba, ignorenx, \
            ctypes.c_void_p(int(bits)), img.width(), img.height(), \
            smooth, ppu, moat, xblend, yblend, zblend)
        node = doc.createNode('Triplanar paste', 'paintlayer')
        node.setOpacity(255)
        doc.rootNode().addChildNode(node, None)
        self.applySelectionAsAlphaMask(layerrgba)
        node.setPixelData(QByteArray(layerrgba.raw), 0, 0, doc.width(), doc.height())
        doc.refreshProjection()

    def createGradient(self):
        doc = self.getDoc()
        if doc == None:
            self.msgBox("Need a texture image open")
            return
        dlg = QDialog(self)
        dlg.setWindowTitle('Create Gradient')
        layout = QFormLayout(dlg)
        box = QVBoxLayout()
        gradientTypeGroup = QButtonGroup(dlg)
        distanceRadioButton = QRadioButton('Distance', dlg)
        box.addWidget(distanceRadioButton)
        gradientTypeGroup.addButton(distanceRadioButton, DISTANCE_GRADIENT)
        directionRadioButton = QRadioButton('Direction', dlg)
        box.addWidget(directionRadioButton)
        gradientTypeGroup.addButton(directionRadioButton, DIRECTION_GRADIENT)
        distanceRadioButton.setChecked(True)
        layout.addRow('Gradient type:', box)
        frame = QFrame(dlg)
        frame.setFrameShape(QFrame.Shape.HLine)
        frame.setFrameShadow(QFrame.Shadow.Sunken)
        layout.addRow(frame)
        box = QVBoxLayout()
        cameraRadioButton = QRadioButton('Camera', dlg)
        box.addWidget(cameraRadioButton)
        topRadioButton = QRadioButton('Top', dlg)
        box.addWidget(topRadioButton)
        bottomRadioButton = QRadioButton('Bottom', dlg)
        box.addWidget(bottomRadioButton)
        leftRadioButton = QRadioButton('Left', dlg)
        box.addWidget(leftRadioButton)
        rightRadioButton = QRadioButton('Right', dlg)
        box.addWidget(rightRadioButton)
        frontRadioButton = QRadioButton('Front', dlg)
        box.addWidget(frontRadioButton)
        backRadioButton = QRadioButton('Back', dlg)
        box.addWidget(backRadioButton)
        cameraRadioButton.setChecked(True)
        layout.addRow('Reference position:', box)
        moatSpinBox = QDoubleSpinBox(dlg)
        moatSpinBox.setValue(4)
        layout.addRow('Moat around geometry:', moatSpinBox)
        ignoreNegativeX = QCheckBox('Ignore negative X (for mirrored objects)')
        layout.addRow(ignoreNegativeX)
        box = QHBoxLayout()
        box.addStretch(1)
        btn = QPushButton('&OK', dlg)
        btn.setDefault(True)
        btn.clicked.connect(dlg.accept)
        box.addWidget(btn)
        btn = QPushButton('&Cancel', dlg)
        btn.clicked.connect(dlg.reject)
        box.addWidget(btn)
        layout.addRow(box)
        if dlg.exec() != QDialog.Accepted:
            dlg.deleteLater()
            return
        dlg.deleteLater()
        gradtype = gradientTypeGroup.checkedId()
        refx = refy = refz = 0.0
        if cameraRadioButton.isChecked():
            model = (ctypes.c_float*16)()
            DLL.modelViewMatrix(ctypes.byref(model))
            refx = self._camd*model[2] + model[3]
            refy = self._camd*model[6] + model[7]
            refz = self._camd*model[10] + model[11]
        elif topRadioButton.isChecked():
            refy = 10000
        elif bottomRadioButton.isChecked():
            refy = -10000
        elif leftRadioButton.isChecked():
            refx = -10000
        elif rightRadioButton.isChecked():
            refx = 10000
        elif frontRadioButton.isChecked():
            refz = 10000
        elif backRadioButton.isChecked():
            refz = -10000
        ignorenx = 1 if ignoreNegativeX.isChecked() else 0
        layerrgba = ctypes.create_string_buffer(doc.width()*doc.height()*4)
        DLL.createGradient(self._mesh, layerrgba, gradtype, ignorenx, ctypes.c_float(moatSpinBox.value()), ctypes.c_float(refx), ctypes.c_float(refy), ctypes.c_float(refz))
        node = doc.createNode('Gradient', 'paintlayer')
        node.setOpacity(255)
        doc.rootNode().addChildNode(node, None)
        self.applySelectionAsAlphaMask(layerrgba)
        node.setPixelData(QByteArray(layerrgba.raw), 0, 0, doc.width(), doc.height())
        doc.refreshProjection()

    def setTextureToCurrentDocument(self):
        self._doc = KI.activeDocument()
        self.updateTexture()
        self.repaint()

    def setTextureToAnyDocument(self):
        self._doc = KI.activeDocument()
        self.updateTexture()
        self.repaint()

    def scalingAt100Percent(self):
        DLL.setScaleDown(1)
        self.updateTexture()
        self.repaint()

    def scalingAt50Percent(self):
        DLL.setScaleDown(2)
        self.updateTexture()
        self.repaint()

    def scalingAt25Percent(self):
        DLL.setScaleDown(4)
        self.updateTexture()
        self.repaint()

    def scalingAt125Percent(self):
        DLL.setScaleDown(8)
        self.updateTexture()
        self.repaint()

    def usePalette(self):
        palettes = list()
        paletteNames = list()
        palettes.append(None)
        paletteNames.append('(none)')
        for (k, p) in KI.resources('palette').items():
            p = Palette(p)
            if p.numberOfEntries() < 1:
                continue
            c = p.colorSetEntryByIndex(0).color()
            if (c.colorDepth() != 'U8') or (c.colorModel() != 'RGBA') or (len(c.components()) != 4):
                continue
            paletteNames.append(k)
            palettes.append(p)
        idx = self.selectItemFromListBox('Use Palette', 'Select palette to use:', paletteNames)
        if idx < 0:
            return
        if idx == 0:
            DLL.setPalette(0, None)
        else:
            p = palettes[idx]
            cols = (ctypes.c_uint32*p.numberOfEntries())()
            for i in range(0, p.numberOfEntries()):
                c = p.colorSetEntryByIndex(i).color()
                b = int(c.components()[0]*255.0)
                g = int(c.components()[1]*255.0)
                r = int(c.components()[2]*255.0)
                a = int(c.components()[3]*255.0)
                rgba = b|(g<<8)|(r<<16)|(a<<24)
                cols[i] = ctypes.c_uint32(rgba)
            DLL.setPalette(p.numberOfEntries(), ctypes.byref(cols))
        self.updateTexture()
        self.repaint()

    def initializeGL(self):
        self._initFailed = False
        try:
            context = self.context()
            context.aboutToBeDestroyed.connect(self.cleanupGL)
            profile = QOpenGLVersionProfile()
            profile.setVersion(2, 1)
            profile.setProfile(QSurfaceFormat.NoProfile)
            gl = self._gl = context.versionFunctions(profile)
            getProcAddressCallbackContext[0] = context
            if DLL.initOpenGL(getProcAddressCallback) != 0:
                self._initFailed = True
        except:
            self._initFailed = True
        getProcAddressCallbackContext[0] = None
        if self._initFailed:
            return
        self.updateTexture()
        gl.glClearColor(self._viewBgColor.redF(), self._viewBgColor.greenF(), self._viewBgColor.blueF(), 1.0)

    def cleanupGL(self):
        if self._initFailed:
            return
        gl = self._gl
        self.context().makeCurrent()
        DLL.unloadAllMeshes()
        DLL.shutdownOpenGL()
        self._mesh = 0

    def doTheRender(self, clearToOpaque):
        gl = self._gl
        if clearToOpaque:
            gl.glClearColor(self._viewBgColor.redF(), self._viewBgColor.greenF(), self._viewBgColor.blueF(), 1.0)
        else:
            gl.glClearColor(self._viewBgColor.redF(), self._viewBgColor.greenF(), self._viewBgColor.blueF(), 0.0)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPushMatrix()
        a = self.width()*1.0/self.height()
        if self._viewOrthographic:
            gl.glOrtho(-self._camd*0.5*a, self._camd*0.5*a, -self._camd*0.5, self._camd*0.5, 0.001, 1000.0)
        else:
            f = 1.0/math.tan(math.radians(45*0.5))
            gl.glLoadMatrixf([f/a, 0, 0, 0,
                0, f, 0, 0,
                0, 0, 1000.001/(0.001 - 1000.0), -1.0,
                0, 0, (2.0*1000.0*0.001)/(0.001 - 1000.0), 0])
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPushMatrix()
        gl.glLoadIdentity()
        gl.glTranslatef(0, 0, -self._camd)
        gl.glRotatef(self._rotx, 1, 0, 0)
        gl.glRotatef(self._roty, 0, 1, 0)
        gl.glTranslatef(-self._tgtx, -self._tgty, -self._tgtz)
        DLL.renderMeshes(ctypes.c_float(self._camd))
        gl.glDisable(gl.GL_LIGHTING)
        gl.glPopMatrix()
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPopMatrix()
        gl.glMatrixMode(gl.GL_MODELVIEW)

    def paintGL(self):
        if self._initFailed:
            return
        self.doTheRender(True)

    def resizeGL(self, width, height):
        if self._initFailed:
            return
        gl = self._gl
        gl.glViewport(0, 0, width, height)


class View3DDocker(DockWidget):

    def __init__(self):
        super().__init__()
        self.setWindowTitle("View 3D")
        self._viewport = View3DViewport()

        layout = QVBoxLayout()
        layout.addWidget(self._viewport)

        buttons = QHBoxLayout()

        chk = QCheckBox("Live")
        chk.setCheckState(Qt.Checked)
        chk.stateChanged.connect(self.liveChanged)
        self._viewport._liveCheckbox = chk
        buttons.addWidget(chk)

        button = QPushButton("Update")
        button.clicked.connect(self._viewport.updateTexture)
        buttons.addWidget(button)

        buttons.addStretch()

        layout.addLayout(buttons)

        w = QWidget(self)
        w.setLayout(layout)
        self.setWidget(w)

    def liveChanged(self, state):
        if self._viewport._initFailed:
            return
        self._viewport._periodicUpdates = state != Qt.Unchecked

    def canvasChanged(self, canvas):
        if self._viewport._initFailed:
            return
        if self._viewport._gl:
            self._viewport.updateTexture()

Krita.instance().addDockWidgetFactory(DockWidgetFactory("view3d", DockWidgetFactoryBase.DockRight, View3DDocker))
